import pytest

from mmuxer.models.condition import All, Any, Body, From, Not, To
from mmuxer.models.enums import ComparisonOperator


def test_condition_base(mailbox, make_message):
    mailbox.login("u", "p")
    make_message(
        content_text="content",
    )
    messages = list(mailbox.fetch())
    m = messages[0]

    cond_ok = From(
        FROM="from@ok.com",
        operator=ComparisonOperator.EQUALS,
    )
    assert cond_ok.eval(m) is True

    comp_cond_not = Not(NOT=cond_ok)
    assert comp_cond_not.eval(m) is False

    cond_ko = To(TO="something else", operator=ComparisonOperator.CONTAINS)
    assert cond_ko.eval(m) is False

    comp_cond_or = Any(ANY=[cond_ok, cond_ko])
    assert comp_cond_or.eval(m) is True

    comp_cond_and = All(ALL=[cond_ok, cond_ko])
    assert comp_cond_and.eval(m) is False


@pytest.mark.parametrize(
    ("content_text", "content_html"),
    (
        ("content", None),
        (None, "<div>content</div>"),
        ("content", "<div>content</div>"),
    ),
)
def test_condition_body(mailbox, make_message, content_text, content_html):
    mailbox.login("u", "p")
    make_message(content_text=content_text, content_html=content_html)
    messages = list(mailbox.fetch())
    m = messages[0]

    cond_ok = Body(BODY="cont")
    assert cond_ok.eval(m) is True

    cond_ko = Body(BODY="contcont")
    assert cond_ko.eval(m) is False



import base64
import imaplib
import json
from pathlib import Path

import pytest

# from imap_tools import BaseMailBox
from mmuxer.mailbox import BaseMailBox


class UnsafeMailBox(BaseMailBox):
    def __init__(self, host, port):
        self._host = host
        self._port = port
        super().__init__()

    def _get_mailbox_client(self) -> imaplib.IMAP4:
        return imaplib.IMAP4(self._host, self._port)


@pytest.fixture
def mailbox():
    box = UnsafeMailBox("localhost", 1143)
    yield UnsafeMailBox("localhost", 1143)
    box._get_mailbox_client().xatom("clear")


@pytest.fixture
def data():
    def inner(filename):
        return (Path(__file__).parent / "data" / filename).read_text()

    return inner


@pytest.fixture
def make_message(mailbox):
    def inner(user="u", box="INBOX", **kwargs):
        if "to" in kwargs:
            assert isinstance(kwargs["to"], list), "`to` must be a list"
        payload = (
            base64.b32encode(
                json.dumps(
                    {
                        **{
                            "to": ["to@ok.com"],
                            "from_": "from@ok.com",
                            "subject": "subject",
                        },
                        **kwargs,
                    }
                ).encode()
            )
            .replace(b"=", b"a")
            .decode()
        )
        mailbox._get_mailbox_client().xatom("receive", user, box, payload)

    return inner



import yaml

from mmuxer.models.rule import Rule


def test_parse_rule(data):
    rules_yaml = data("models/rules.yaml")
    rules_dicts = yaml.safe_load(rules_yaml)
    for f in rules_dicts:
        r = Rule.model_validate(f)
        r._actions()



import pytest

from mmuxer.models.condition import All, Any, From, Not
from mmuxer.models.sieve import parse_condition, to_condition, to_dnf, to_sieve_conditions

cond1 = From(FROM="1")
cond2 = From(FROM="2")
cond3 = From(FROM="3")


@pytest.mark.parametrize(
    "condition",
    (
        cond1,
        cond2,
        Not(NOT=cond1),
        All(ALL=[cond1, Not(NOT=cond2)]),
        Any(ANY=[cond1, Not(NOT=cond2)]),
        Not(NOT=Any(ANY=[cond1, Not(NOT=cond2)])),
    ),
)
def test_convert_condition(condition):
    expr = parse_condition(condition)
    assert to_condition(expr) == condition


@pytest.mark.parametrize(
    "condition, expected",
    (
        (cond1, cond1),
        (Not(NOT=cond1), Not(NOT=cond1)),
        (Any(ANY=[cond1, cond2]), Any(ANY=[cond1, cond2])),
        (Any(ANY=[cond1, Not(NOT=cond2)]), Any(ANY=[cond1, Not(NOT=cond2)])),
        (All(ALL=[cond1, Not(NOT=cond2)]), All(ALL=[cond1, Not(NOT=cond2)])),
        (Any(ANY=[cond1, All(ALL=[cond2, cond3])]), Any(ANY=[cond1, All(ALL=[cond2, cond3])])),
        (
            All(ALL=[cond1, Any(ANY=[cond2, cond3])]),
            Any(
                ANY=[
                    All(ALL=[cond1, cond2]),
                    All(ALL=[cond1, cond3]),
                ]
            ),
        ),
        (
            Not(NOT=All(ALL=[cond1, Any(ANY=[cond2, cond3])])),
            Any(
                ANY=[
                    Not(NOT=cond1),
                    All(ALL=[Not(NOT=cond2), Not(NOT=cond3)]),
                ]
            ),
        ),
        (
            All(ALL=[cond1, Not(NOT=cond2), All(ALL=[cond3])]),
            All(ALL=[cond1, Not(NOT=cond2), cond3]),
        ),
        (
            All(ALL=[cond1, Not(NOT=cond2), All(ALL=[cond3, cond3])]),
            All(ALL=[cond1, Not(NOT=cond2), cond3]),
        ),
    ),
)
def test_to_dnf(condition, expected):
    dnf = to_dnf(condition)
    assert dnf == expected


@pytest.mark.parametrize(
    "condition",
    (
        (cond1),
        (Not(NOT=cond1)),
        (Any(ANY=[cond1, cond2])),
        (Any(ANY=[cond1, Not(NOT=cond2)])),
        (All(ALL=[cond1, Not(NOT=cond2)])),
        (Any(ANY=[cond1, All(ALL=[cond2, cond3])])),
        (All(ALL=[cond1, Any(ANY=[cond2, cond3])])),
        (Not(NOT=All(ALL=[cond1, Any(ANY=[cond2, cond3])]))),
        (All(ALL=[cond1, Not(NOT=cond2), All(ALL=[cond3])])),
        (All(ALL=[cond1, Not(NOT=cond2), All(ALL=[cond3, cond3])])),
    ),
)
def test_to_sieve_rules(condition):
    to_sieve_conditions(condition)



"""
Code modified from https://github.com/ikvk/imap_tools/blob/master/imap_tools/mailbox.py
"""

import imaplib
import sys
from itertools import islice
from typing import Iterable, Iterator, Sequence

from imap_tools.errors import MailboxFetchError
from imap_tools.mailbox import BaseMailBox as BaseMailBoxOG
from imap_tools.mailbox import check_timeout_arg_support
from imap_tools.utils import check_command_status, chunks

PYTHON_VERSION_MINOR = sys.version_info.minor


def batched(iterable: Iterable, n: int):
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


class BaseMailBox(BaseMailBoxOG):

    def _fetch_in_bulk(
        self, uid_list: Sequence[str], message_parts: str, reverse: bool
    ) -> Iterator[list]:
        if not uid_list:
            return
        # batches of size 100 seem like a sweet spot in terms of perf, and
        # should prevent any issue with requests too big
        batches = batched(uid_list, 100)
        for uid_batch in batches:
            fetch_result = self.client.uid("fetch", ",".join(uid_batch), message_parts)
            check_command_status(fetch_result, MailboxFetchError)
            if not fetch_result[1] or fetch_result[1][0] is None:
                return
            for built_fetch_item in chunks((reversed if reverse else iter)(fetch_result[1]), 2):
                yield built_fetch_item


class MailBox(BaseMailBox):
    """Working with the email box through IMAP4 over SSL connection"""

    def __init__(
        self, host="", port=993, timeout=None, keyfile=None, certfile=None, ssl_context=None
    ):
        """
        :param host: host's name (default: localhost)
        :param port: port number
        :param timeout: timeout in seconds for the connection attempt, since python 3.9
        :param keyfile: PEM formatted file that contains your private key (deprecated)
        :param certfile: PEM formatted certificate chain file (deprecated)
        :param ssl_context: SSLContext object that contains your certificate chain and private key
        Since Python 3.9 timeout argument added
        Since Python 3.12 keyfile and certfile arguments are deprecated, ssl_context and timeout must be keyword args
        """
        check_timeout_arg_support(timeout)
        self._host = host
        self._port = port
        self._timeout = timeout
        self._keyfile = keyfile
        self._certfile = certfile
        self._ssl_context = ssl_context
        super().__init__()

    def _get_mailbox_client(self) -> imaplib.IMAP4:
        if PYTHON_VERSION_MINOR < 9:
            return imaplib.IMAP4_SSL(
                self._host, self._port, self._keyfile, self._certfile, self._ssl_context
            )
        elif PYTHON_VERSION_MINOR < 12:
            return imaplib.IMAP4_SSL(
                self._host,
                self._port,
                self._keyfile,
                self._certfile,
                self._ssl_context,
                self._timeout,
            )
        else:
            return imaplib.IMAP4_SSL(
                self._host, self._port, ssl_context=self._ssl_context, timeout=self._timeout
            )
